/* Copyright 2008-2013 Research Foundation State University of New York   */

/* This file is part of QUB Express.                                      */

/* QUB Express is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Express is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Express program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

#ifndef QUBX_MACRO_R_H
#define QUBX_MACRO_R_H

#include "qubxmacror.h"
#include "../qubfast/qubx_model.h"
#include "callbk_reportfun.h"

#ifdef __cplusplus
extern "C" {
#endif

/*
 begin_html
 <h3>Maximum Macroscopic Likelihood</h3>
 [from (Milescu 2005)] by Chris Nicolai 2009
 <p>
 This function computes the log-likelihood of data given a model.
 </p><p>
 See also:
 <ul>
 <li><a href="max_mac_ll.cpp.html">max_mac_ll.cpp</a></li>
 <li><a href="../qubfast/max_ll_util.h.html">common/max_ll_util.h</a></li>
 <li><a href="../common/callbk_reportfun.h.html">common/callbk_reportfun.h</a></li>
 </ul>
 <b>Up:</b> <a href="index.html">Index</a>
 </p>

 <h4>The Model</h4>
 <p>
 Our Markov model is a graph with colored vertices.  A vertex is called a "state,"
 and its color is a nonnegative integer called its "class."  States in the same
 class are indistinguishable (same amp).  To describe the vertices, provide the array
 </p><pre>
    int clazz[Ns] = [class of each state]
    int Ns        = number of states
 </pre><p>
 Each edge is labeled with its transition rate (probability per second).  These form the matrix
 </p><p>
    Q, a Ns x Ns matrix with<br>
    \(Q_{a,b}\) = rate from state a to state b<br>
    \(Q_{a,a} = - \sum_i Q_{a,i}\) where \(i \neq a\)
 </p><p>
 Each \(Q_{a,b} = K0_{a,b} * Ligand_{a,b} * e^{K1_{a,b} * Voltage_{a,b}}\).  You provide the Ns x Ns matrices
 </p><pre>
    double **K0, **K1            of kinetic parameters
    int    **Ligand, **Voltage   index of the ligand or voltage signal influencing each rate, or 0
 </pre><p>
 with the diagonals undefined.
 </p><p>
 A pair of states is either connected in both directions or neither.  To indicate un-connectedness, set
 </p><p>
    \(K0_{a,b} = K0_{b,a} = 0\)
 </pre>
 <p>
 For constant state entry probabilities, as described in (Qin...1996), provide the array
 </p><pre>
 int P0[Ns]
 </pre><p>
 For equilibrium entry probabilities, as described in (Milescu...2005), provide
 </p><pre>
 P0 = NULL
 </pre>

TODO: talk about amp/cond model and channel count


 <h4>The Data</h4>
 <p>                            This is wrong -- leftover from MSL -- 

 The data consist of one or more parallel signals.  The first (index 0) is the Markovian
 one to be analyzed.  Each additional signal describes a ligand or voltage variable.
 Signals are in model order; i.e., if (v = Voltage[a][b]) != 0, then 
 the signal at index v is the voltage controlling a->b.
 For each signal you provide
 </p><pre>
 int    DwellCount            number of events
 int    Classes[DwellCount]   class index of each event
 float  Durations[DwellCount] length of each dwell in milliseconds
 double Amps[ClassCount]      amplitude (or ligand/voltage value) of each class
 </pre><p>
 and you give the signals together as
 </p><pre>
 int    Nsignal                  number of signals
 int    *dwellCounts
 int    **classeses
 float  **durationses
 double **ampses
 </pre><p>
 You provide one or more segments of data, each with the same number of signals.
 All together they are given as
 </p><pre>
 int    Nseg
 int    Nsignal
 int    **dwellCountses          [segment][signal]
 int    ***classeseses
 float  ***durationseses
 double ***ampseses
 </pre><p>
 We multiplex the signals to create a single idealized signal which changes class
 whenever any source signal changes class.  Each plexi-class denotes a Markov class
 with a specific set of experimental constants (the other signals' idealized amplitudes).
 </p>
 Then, as in MIL, we subtract tdead.
 tdead is the longest duration of events that can't be reliably detected.
 MSL deletes any such events, by merging them with their prior.  Then
 for computational reasons, tdead is subtracted from each event, and they're converted to seconds.
 </p><p>
 Sadly, someone has to allocate memory to store this processed signal.  It has the form
 </p><pre>
 int     Nseg
 int     Nsignal
 int     dwellCounts[Nseg]
 int     **classeses
 float   **durationses
 int     Nplex
 int     plexiclasses[2*Nplex]          alternating markov-class[i/2], stimclass[i/2]
 int     Nstim
 double  stimclasses[Nstim*Nsignal]     [stimcls * Nsignal + signal_ix], signal 0 undefined
 </pre><p>
 First, we can scan your data to compute upper bounds for dwellCounts, Nplex and Nstim.
 You allocate dwellCounts[Nseg] and call:
 </p>
 end_html
 */

// QUBXMACROR_API void __stdcall mac_multiplex_bounds(int Nseg, int Nsignal,
//							  int **dwellCountses, int ***classeseses,
//							  float ***durationseses, double ***ampseses,
//							  int *out_dwellCounts, int *out_Nplex, int *out_Nstim);
//
/*
begin_html
<p>
Then you allocate classeses[Nseg][dwellCounts[i]] and durationses, plexiclasses and stimclasses,
and call this function to multiplex and process the signals:
</p>
end_html
*/

// QUBXMACROR_API void __stdcall mac_multiplex(int Nseg, int Nsignal,
//						   int **dwellCountses, int ***classeseses,
//						   float ***durationseses, double ***ampseses,
//						   double tdead_ms, int *out_dwellCounts,
//						   int **out_classeses, float **out_durationses,
//						   int *out_Nplex, int *out_plexiclasses,
//						   int *out_Nstim, double *out_stimclasses);




typedef struct
{
  float time, cur, var;
} qubx_macror_sample;

typedef struct
{
  float fit, fit_var, ll;
} qubx_macror_sample_out;


// opaque type qubx_macror_data = (void *):
//   stores data (uneven sampling possible, or no data) and stimulus idealization
QUBXMACROR_API void* __stdcall qubx_macror_data_create(int Nstimsig, int data_cap);
QUBXMACROR_API void __stdcall qubx_macror_data_reset(void *data, int Nstimsig, int data_cap);
QUBXMACROR_API void __stdcall qubx_macror_data_free(void *data);
QUBXMACROR_API int __stdcall qubx_macror_data_append_segment(void *data);
QUBXMACROR_API void __stdcall qubx_macror_data_append_chunk(void *data, int seg_ix, double dur_sec, int Nd,
							    float *T, float *I, float *V, int *dwellCounts,
							    int **classeses, float **durationses, double **ampses, double even_sampling_sec);
  // if evenly sampled, provide sampling rate, else even_sampling_sec=0.0;

QUBXMACROR_API double* __stdcall qubx_macror_data_get_stimclasses(void *data, int *Nstimclass);
// returns double[Nstimclass * Nsignal], Nsignal = Nstimsig + 1;
QUBXMACROR_API double* __stdcall qubx_macror_data_get_stimclass_frac(void *data); // after get_stimclasses
// returns double[Nstimclass]; weight of each stimclass; sums to 1.0; valid after calling get_stimclasses()
QUBXMACROR_API qubx_macror_sample_out* __stdcall qubx_macror_data_get_buffer(void *data);

typedef struct {
  void *storage;
  
  // you set:
  void *data;
  int useVar;
  int optNchannel;
  int accel;

  // and add_model() -- exactly once
  
  // ready after setup_constraints():
  int Npar, NfPar;
  double *pars, *fPars;

  // to gray out the GPU checkbox
  int gpu_fail;

  // output after qubx_macror_ll()
  double ll;
  
  // this is what's actually used (initialized from fastmodel nchannel)
  double float_Nchannel;
} qubx_macror_param;

QUBXMACROR_API qubx_macror_param* __stdcall qubx_macror_param_create();
QUBXMACROR_API void __stdcall qubx_macror_param_free(qubx_macror_param *p);

  // TODO: is it possible to incorporate multiple (summed) source models into the recursive calculation? (specif. the posterior state distribution)
  // for now, exactly one model may be added.
QUBXMACROR_API void __stdcall qubx_macror_param_add_model(qubx_macror_param *p, qubx_model *model, qubx_stim_amps *ampm, qubx_stim_rates *ratesm);
QUBXMACROR_API int  __stdcall qubx_macror_setup_constraints(qubx_macror_param *p);
QUBXMACROR_API void __stdcall qubx_macror_do_fPars_to_model(qubx_macror_param *p);
QUBXMACROR_API int __stdcall qubx_macror_ll(qubx_macror_param *p);
QUBXMACROR_API void __stdcall qubx_macror_calc_std(qubx_macror_param *p, double *InvHessian);




#ifdef __cplusplus
}
#endif

#endif

