"""Base classes for QUB Express plugins, and sample implementations.

This unit is not required to make a plugin (see qubx.pyenv),
but is intended to simplify and explain the process.

To make a plugin:
    * make a folder in qub-express's Plugins folder (e.g. "my_plugin")
    * make a subfolder with a unique name for your plugin's package (e.g. "my_plugin/my_uniquestr_package");
      the my_plugin folder will be part of sys.path; if there is any other my_uniquestr_package in sys.path, it
      may be loaded instead.
    * copy this file into my_uniquestr_package; by putting it there, you're guaranteed to import this copy,
      and not one from some other plugin.
    * make a python module, e.g. "my_uniquestr_package/my_module.py", which starts with "from . import qubx_plugin",
      and defines a subclass of one of the plugin base classes.
    * make a python module, "my_uniquestr_package/__init__.py" along the lines of:
__all__ = ['qubx_plugin', 'my_module']
      or if you want, leave it blank, but make sure it exists.
    * make these text files in the my_plugin folder:
       - qubx_plugin_about.txt   - information for the Admin:Plugins panel
       - qubx_plugin_imports.txt - newline-separated list of modules, in the order they should be re-loaded
       - qubx_plugin_init.py     - python script executed on load;
                                   typically creates and adds one or more "Face" to the tool/about panels.
       - qubx_plugin_fini.py     - python script executed on unload; typically removes the "Face"(s)
       - qubx_plugin_name.txt    - name of plugin, as shown in Admin:Plugins panel

For examples, see qubx_sample_plugin, qubx_macroscopic, and qubx_single_molecule.
"""

import gobject
import gtk
import pango
import traceback

import qubx.notebook
import qubx.optimize
import qubx.optimizeGTK
import qubx.pyenv
import qubx.trialset
from qubx.accept import *
from qubx.faces import Face
from qubx.GTK import Requestable, pack_item, pack_space, pack_label, pack_scrolled, pack_check, pack_button
from qubx.model import QubModel, ModelToQ, ModelToP0, ModelToClazz, QtoPe, K012ToModel, ModelToRateMatricesP
from qubx.task import Robot, Tasks
from qubx.util_types import Anon, Product, Reffer, WeakEvent

from itertools import izip, count
from numpy import array, matrix, zeros, any, sum
import scipy.linalg.matfuncs
expm = scipy.linalg.matfuncs.expm

class AboutPlugin(Face):
    def __init__(self, label, global_name=""):
        super(AboutPlugin, self).__init__(label, global_name=global_name)

class ToolPlugin(Face):
    """Base class for top-pane plugin interface with optional worker thread (robot).

    @ivar robot: L{qubx.task.Robot} managing a background work thread
"""

    # these names are visible in the Explore dialog (Admin:Scripts):
    __explore_featured = ['robot', 'on_robot_begin', 'on_robot_end', 'on_exception', 'on_interrupt']

    def __init__(self, label, global_name="", has_robot=True):
        """
        @param label: name on this tab in the gui
        @param has_robot: if True, starts a background thread called self.robot
"""
        super(ToolPlugin, self).__init__(label, global_name=global_name)
        self.__ref = Reffer()
        self.set_size_request(200, 50)
        
        if has_robot:
            self.robot = Robot(label, self.__ref(self.__on_robot_begin), self.__ref(self.__on_robot_end))
            self.robot.OnException += self.__ref(self.on_exception)
            self.robot.OnInterrupt += self.__ref(self.on_interrupt)
            QubX = qubx.pyenv.env.globals['QubX']
            QubX.OnQuit += self.__ref(self.__qubx_finalize)
        else:
            self.robot = None
    def __qubx_finalize(self):
        self.robot.stop()
    def __on_robot_begin(self):
        Tasks.add_task(self.robot)
        self.on_robot_begin()
    def __on_robot_end(self):
        Tasks.remove_task(self.robot)
        self.on_robot_end()
    def on_robot_begin(self):
        """Override this method to respond when the robot starts working."""
        pass
    def on_robot_end(self):
        """Override this method to respond when the robot finishes working."""
        pass
    def on_exception(self, task, typ, val, tb):
        """Override this method to do something with exceptions thrown on robot thread; default: print to stdout."""
        traceback.print_exception(typ, val, tb)
    def on_interrupt(self, task, cancel):
        """Override this method to customize the effect of the Stop button (e.g. set a flag); default: KeyboardInterrupt raised in robot thread."""
        # if you don't call cancel, qubx will throw a KeyboardInterrupt into the robot thread.
        pass

class OptToolPlugin(ToolPlugin):
    """Base class for optimization plugins.

Override at minimum: robot_calculate_score, robot_get_pars, robot_set_pars.
Optionally, override robot_optimize to replace the built-in optimizers.

They are called robot_* because they run on the worker (robot) thread.  When called,
the latest model has been copied into self.robot.model, a L{QubModel}.

Add any additional widgets to self.top_panel and self.right_panel, or rearrange layout as needed.

model_prep has .source (the QubModel in GUI), .model (private copy for robot computation),
and .set_rates2(K0, K1, K2) [faster than making new model_prep]

    @ivar right_panel: L{gtk.VBox} along right edge, with optimize controls already at the bottom
    @ivar top_panel: L{gtk.HBox} along top edge, with score display already at far right
    @ivar txtStatus: L{gtk.TextView} with text output; convenience methods self.append_output(s), self.clear_output()
    @ivar model_prep: L{ModelPrep} with source model and private copy for robot use
    
    """

    __explore_featured = ['right_panel', 'top_panel', 'txtStatus', 'model_prep', 'grads', 'hessian',
                          'OnLL', 'OnIteration', 'OnOptimized', 'model_tracker', 'data', 'data_prep',
                          'btnGetScore', 'txtScore', 'lblScore', 'btnAddToTrials', 'txtStatus',
                          'btnOpt', 'optimizing', 'score', 'model', 'get_model_prep', 'get_data_prep', 'request_data_prep',
                          'append_output', 'clear_output', 'add_to_trials', 'make_trial', 'get_ndata', 'add_trial_fields',
                          'calculate_score', 'optimize', 'robot_setup_data', 'update_score', 'robot_calculate_score',
                          'robot_get_pars', 'robot_set_pars', 'robot_optimize']

    def __init__(self, label, global_name="", opt_name=None, opt_options={}, have_grads=False):
        """
        @param label: tab label, becomes self.face_name
        @param opt_name: optimizer preferences selector, defaults to label
        @param opt_options: {option_label : default} to override specific defaults of the built-in optimizers (see qubx.optimize)
        @param have_grads: True if your robot_calculate_score can also calculate dScore/dPar[i]
"""
        super(OptToolPlugin, self).__init__(label, global_name=global_name)
        self.__ref = Reffer()
        self.__score = 0.0
        self.__optimizing = False
        self.have_grads = have_grads
        self.opt_name = opt_name or label
        qubx.optimize.setup_opt(self.opt_name, opt_options)

        self.grads = []
        self.hessian = None
        self.OnLL = WeakEvent()
        self.OnIteration = WeakEvent()
        self.OnOptimized = WeakEvent()
        
        self.model_tracker = ModelTracker()
        self.model_tracker.OnSwitch += self.__ref(self.__onSwitchModel)
        self.model_tracker.OnChange += self.__ref(self.__onSetModel)
        self.model_tracker.OnSetRate += self.__ref(self.__onSetRate)
        self.model_prep = Product(self.get_model_prep)
        self.model_prep.OnInvalidate += self.__ref(self.__onInvalidateModel)
        self.rates = Product(self.__get_rates)
        self.rates.OnInvalidate += self.__ref(self.__onInvalidateModel)
        self.__model = None
        self.data = OptToolData(self.get_data_prep)
        self.data.OnInvalidate += self.__ref(self.__onInvalidateData)
        self.data_prep = None

        self.top_panel = pack_item(gtk.HBox(), self)
        v = pack_item(gtk.VBox(), self.top_panel, at_end=True)
        vh = pack_item(gtk.HBox(), v)
        self.btnGetScore = pack_button('Get Score', vh, at_end=True, on_click=self.__onClickGetScore)
        self.txtScore = pack_item(gtk.Entry(), vh, at_end=True)
        self.txtScore.set_width_chars(10)
        self.lblScore = pack_label('Score:', vh, at_end=True)
        self.btnAddToTrials = pack_button('Add to Trials', v, expand=True, on_click=self.__onClickAddToTrials)
        pack_space(self.top_panel, x=20, at_end=True)
        
        box = pack_item(gtk.HBox(), self, expand=True)
        self.txtStatus = gtk.TextView()
        self.txtStatus.set_editable(False)
        self.txtStatus.modify_font(pango.FontDescription('monospace 10'))
        pack_scrolled(self.txtStatus, box, expand=True)
        
        self.right_panel = pack_item(gtk.VBox(), box)
        h = pack_item(gtk.HBox(True), self.right_panel, at_end=True)
        self.btnOpt = pack_button('Optimize', h, on_click=self.__onClickOptimize)
        
        self.optOptions = qubx.optimizeGTK.OptOptionsFace(label, self.opt_name, 'QubX.About.%s'%label)
        # set self.optOptions to None if you're not using qubx.optimize
        
        try:
            qubx.global_namespace.QubX.Modeling.Utils.add_optimizer(label, self.get_model_prep, self.get_data_prep,
                                                                    self.calculate_score, self.optimize,
                                                                    self.OnLL, self.OnIteration, self.OnOptimized)
        except:
            traceback.print_exc()
        
        self.__onSwitchModel(qubx.global_namespace.QubX.Models.file)
    optimizing = property(lambda self: self.__optimizing)
    def set_score(self, x):
        gobject.idle_add(self.txtScore.set_text, "%.5g" % x)
        self.__score = x
    score = property(lambda self: self.__score, lambda self, x: self.set_score(x),
                     "quantity to maximize, e.g. log-likelihoo; auto-displayed at top-right")
    def onShow(self, showing):
        # if using built-in optimizers, show their options in the About panel
        if not self.optOptions:
            return
        QubX = qubx.pyenv.env.globals['QubX']
        if showing:
            QubX.About.append_face(self.optOptions)
            QubX.About.show_face(self.optOptions.face_name)
        else:
            QubX.About.remove_face(QubX.About.index(self.optOptions.face_name))
    def __onSwitchModel(self, model):
        if self.__optimizing: # put it off until optimization finishes
            self.robot.do(gobject.idle_add, self.__onSwitchModel, model)
        else:
            self.__model = model
            self.__onSetModel()
    model = property(lambda self: self.__model)
    def __onInvalidateModel(self, product):
        pass
        #if not self.__optimizing:
        #    self.robot.do(gobject.idle_add, self.invalidate_controls)
    def __onSetModel(self, *args):
        self.model_prep.invalidate()
        self.rates.invalidate()
    def __onSetRate(self, r, field, val, prev, undoing):
        self.rates.invalidate()
    def get_model_prep(self, model=None):
        # override to customize the ModelPrep
        m = model or self.__model
        if not m:
            return None
        return ModelPrep(m)
    def __onInvalidateData(self, product):
        self.model_prep.invalidate()
        self.rates.invalidate()
    def get_data_prep(self, segs, model_prep, wait=True, receiver=lambda dp: None):
        # leave this one alone; override the next one to customize DataPrep
        if wait:
            return qubx.pyenv.call_async_wait(self.request_data_prep, segs, model_prep, last_data_prep=self.data_prep)
        else:
            self.request_data_prep(receiver, segs, model_prep, last_data_prep=self.data_prep)
    def request_data_prep(self, receiver, segs=None, model_prep=None, last_data_prep=None):
        # override to customize the DataPrep
        data_prep = Anon(segs=segs, model_prep=model_prep)
        receiver(data_prep)
    def __get_rates(self):
        if self.__model == None:
            return None, None
        signal_names = qubx.global_namespace.QubX.Modeling.Stimulus.active_names
        K0, K1, K2, L, V, P = ModelToRateMatricesP(self.__model, signal_names)
        return K0, K1, K2
    def on_robot_begin(self):
        def do_on_idle():
            self.btnGetScore.set_sensitive(False)
            self.btnOpt.set_sensitive(False)
        gobject.idle_add(do_on_idle)
    def on_robot_end(self):
        def do_on_idle():
            self.btnGetScore.set_sensitive(True)
            self.btnOpt.set_sensitive(True)
        gobject.idle_add(do_on_idle)
    def __onClickGetScore(self, btn):
        qubx.pyenv.env.scriptable_if_matching('%s.calculate_score()' % self.global_name,
                                              [(self.global_name, self)])
        self.calculate_score(wait=False, receiver=self.set_score)
    def __onClickAddToTrials(self, btn):
        self.add_to_trials(*self.make_trial())
    def __onClickOptimize(self, btn):
        qubx.pyenv.env.scriptable_if_matching('%s.optimize(on_trial=%s.add_to_trials)' % (self.global_name, self.global_name),
                                              [(self.global_name, self)])
        self.__mtree_pre_opt = self.__model.as_tree()
        self.optimize(wait=False, on_trial=self.add_to_trials)
    def append_output(self, s):
        """Appends a string to the output panel."""
        buf = self.txtStatus.get_buffer()
        buf.insert(buf.get_end_iter(), s)
        self.txtStatus.scroll_to_mark(buf.get_insert(), 0)
    def clear_output(self):
        """Clears the text output panel."""
        self.txtStatus.set_text("")
    def on_exception(self, task, typ, val, tb):
        self.append_output('\n'.join(traceback.format_exception(typ, val, tb)))
    def add_to_trials(self, mtree, row):
        qubx.pyenv.env.OnScriptable('%s.add_to_trials(*%s.make_trial())' % (self.global_name, self.global_name))
        Trials = qubx.global_namespace.QubX.Trials
        ts = Trials.get_trialset(self.face_name, in_qsf=True)
        ts.append_trial(mtree, row)
        Trials.trialset = ts
    def make_trial(self, model_prep=None, data_prep=None, fields={}, initial_tree=None):
        try:
            QubX = qubx.global_namespace.QubX
            if model_prep:
                model = model_prep.source or model_prep.model
            else:
                model = self.QubX.Models.file
            if data_prep and data_prep.segs:
                data = data_prep.segs[0].file
            else:
                data = self.QubX.Data.file
            mtree = model.as_tree()
            nParam = fields['nParam'] if ('nParam' in fields) else 0
            if 'nData' in fields:
                nData = fields['nData']
            else:
                nData = self.get_ndata(data_prep)
            row = qubx.trialset.make_trial_row(model.path, data.path, self.score, nParam, nData=nData)
            row['nData'] = nData
            if initial_tree:
                self.add_trial_fields(row, initial_tree, 'Initial ', model_prep, data_prep)
            self.add_trial_fields(row, mtree, "", model_prep, data_prep)
            for k,v in fields.iteritems():
                row[k] = v
            return mtree, row
        except:
            traceback.print_exc()
    def get_ndata(self, data_prep=None):
        """Override to tell how many datapoints/events used, for Trials."""
        return 0
    def add_trial_fields(self, row, mtree, prefix='', model_prep=None, data_prep=None):
        """Override to pick info from the model."""
        pass # e.g. qubx.trialset.add_trial_rates(row, mtree, prefix)
    def calculate_score(self, model_prep=None, data_prep=None, wait=True, receiver=lambda score: None, on_trial=None, **kw):
        """Starts the robot calculating the score.
        @param receiver: func(score) called when score is finished calculating
"""
        if wait:
            qubx.pyenv.call_async_wait(self.__calculate_score, model_prep, data_prep, on_trial=on_trial, **kw)
            receiver(self.__score)
            return self.__score
        else:
            self.__calculate_score(receiver, model_prep, data_prep, on_trial=on_trial, **kw)
    def __calculate_score(self, receiver=lambda score: None, model_prep=None, data_prep=None, on_trial=None, **kw):
        QubX = qubx.pyenv.env.globals['QubX']
        self.__set_keys(**kw)
        mp = model_prep
        if mp is None:
            mp = self.model_prep.val
            mp.set_rates2(*self.rates.val)
        self.robot.do(self.__robot_calculate_score, receiver, mp, data_prep, on_trial)
    def optimize(self, model_prep=None, data_prep=None, wait=True, receiver=None, ext_return=False, on_trial=None, **kw):
        """Starts the robot optimizing.
        @param receiver: func() called when optimization is finished
"""
        if wait:
            qubx.pyenv.call_async_wait(self.__optimize, model_prep, data_prep, on_trial=on_trial, **kw)
            if ext_return:
                if receiver:
                    receiver(self.score, self.iterations, self.grads, self.hessian)
                return self.score, self.iterations, self.grads, self.hessian
            else:
                if receiver:
                    receiver(self.score)
                return self.score
        else:
            self.__optimize(receiver, model_prep, data_prep, ext_return, on_trial=on_trial, **kw)
    def __optimize(self, receiver=None, model_prep=None, data_prep=None, ext_return=False, on_trial=None, **kw):
        QubX = qubx.pyenv.env.globals['QubX']
        self.__set_keys(**kw)
        mp = model_prep
        if mp is None:
            mp = self.model_prep.val
            if mp.source:
                mp.source.undoStack.enable(False)
            mp.set_rates2(*self.rates.val)
        self.__mtree_pre_opt = mp.model.as_tree()
        QubX.Trials.get_trialset("%s Iterations"%self.face_name, in_qsf=True).clear()
        self.__optimizing = True
        self.robot.do(self.__robot_optimize, receiver, mp, data_prep, ext_return, on_trial)
    def __set_keys(self, **kw):
        for k, v in kw.iteritems():
            try:
                setattr(self, k, v)
            except:
                print 'MacRates optimizer has no option named %s' % k
    def robot_setup_data(self, model_prep, data_prep):
        self.robot.model_prep = model_prep
        if data_prep:
            self.data_prep = data_prep
        else:
            self.data.model_prep = model_prep
            self.data_prep = self.robot.gui_call_recv(self.data.request, self.robot)[0]
    def __robot_calculate_score(self, on_finish=lambda: None, model_prep=None, data_prep=None, on_trial=None):
        """Calculates score on worker thread, then passes it back to GUI thread."""
        self.robot_setup_data(model_prep, data_prep)
        try:
            score = self.robot_calculate_score()
        except:
            score = 0.0
            self.robot.send_exception()
        if on_trial:
            mtree, row = self.make_trial(model_prep, data_prep)
            row['LL'] = score
            gobject.idle_add(on_trial, mtree, row)
        gobject.idle_add(self.update_score, score, lambda: on_finish(score))
        gobject.idle_add(self.OnLL, score)
    def __robot_optimize(self, on_finish=None, model_prep=None, data_prep=None, ext_return=False, on_trial=None):
        """Optimizes on worker thread, then notifies GUI thread."""
        self.robot_setup_data(model_prep, data_prep)
        mtree_initial = model_prep.model.as_tree()
        try:
            score, pars, iterations, grads, Hessian = self.robot_optimize()
            self.iterations = iterations
            self.grads = grads
            self.hessian = Hessian
        except:
            self.robot.send_exception()
            score = 0.0
        self.score = score
        self.__optimizing = False
        if on_trial:
            row = {'nParam' : len(self.grads),
                   'Iterations' : self.iterations}
            for i, g in enumerate(self.grads):
                row['Gradient %i' % i] = g
            gobject.idle_add(on_trial, *self.make_trial(model_prep, data_prep, row, mtree_initial))
        if ext_return:
            gobject.idle_add(self.update_score, score, lambda: on_finish and on_finish((score, iterations, grads, Hessian)))
        else:
            gobject.idle_add(self.update_score, score, lambda: on_finish and on_finish(score))
        gobject.idle_add(self.__onOptimized, iterations, score, grads, Hessian)
    def __onOptimized(self, iterations, score, grads, Hessian):
        model_prep = self.robot.model_prep
        if model_prep.source:
            model_prep.source.undoStack.enable()
            model_prep.source.undoStack.push_undo(bind(model_prep.source.from_tree, self.__mtree_pre_opt),
                                                  bind(model_prep.source.from_tree, model_prep.source.as_tree()))
            model_prep.source.undoStack.seal_undo('set rates')
        ts = qubx.global_namespace.QubX.Trials.get_trialset("%s Iterations"%self.face_name, in_qsf=True)
        for mtree, row in self.iter_trials:
            ts.append_trial(mtree, row)
        self.OnOptimized(iterations, score, grads, Hessian)
    def update_score(self, score, call_me=lambda: None):
        """Updates score and txtScore then calls call_me().  Intended as GUI callback from robot thread; e.g.
    >>> gobject.idle_add(self.update_score, score, on_finish)
"""
        self.score = score
        call_me()
    def robot_calculate_score(self, grads=None):
        """Calculates and returns the score, on worker thread; override this method.
self.model_prep.model has just been updated.
        @param grads: either None (no gradient requested) or an array to fill with dScore/dPar[i]
"""
        if grads != None:
            grads[0] = 0.0
            grads[1] = 0.0
        return 1.0
    def robot_get_pars(self):
        """Returns the list of initial parameter values, on worker thread; override this method.
Derive parameter values from self.robot.model_prep.model and/or from your own widgets."""
        return [1.0, -1.0]
    def robot_set_pars(self, pars):
        """Updates self.robot.model_prep.model and/or other state from new param values, on worker thread; override this method."""
        pass
    def robot_optimize(self):
        """Finds improved param values, starting from robot_get_pars(); override if you don't want the built-in optimizers.
If you override, also do self.optOptions = None.
self.robot.model_prep.model has just been updated.
"""
        pars = self.robot_get_pars()
        self.robot.iter = 0
        self.iter_trials = []
        self.robot.maxiter = qubx.optimize.get_options(self.opt_name)['max iterations']
        if self.optOptions:
            optimize = qubx.optimize.get_opts(self.opt_name,
                                              on_start=lambda rec, opts: gobject.idle_add(self.optOptions.optimizers.onStartOpt, rec, opts),
                                              on_end=lambda rec, opts: gobject.idle_add(self.optOptions.optimizers.onEndOpt, rec, opts))
        else:
            optimize = qubx.optimize.get_opts(self.opt_name)
        score, pars, iterations, grads, Hessian = optimize(array(pars), self.__robot_func, self.__robot_iter, self.have_grads and self.__robot_grad_func or None)
        # important: use idle_add to pass GUI-changing stuff to the main thread:
        gobject.idle_add(self.append_output, "Optimization finished after %d iterations.\nScore = %.9g\n" % (iterations, -score))
        #   (on linux, it can be any thread, as long as the main thread is paused [with robot.main_hold:],
        #    but windows will crash unless all GUI API calls come from one thread.)
        
        return -score, pars, iterations, grads, Hessian
    # callbacks for built-in optimizers:
    def __robot_func(self, pars):
        self.robot_set_pars(pars)
        return - self.robot_calculate_score()
    def __robot_grad_func(self, pars, grads):
        self.robot_set_pars(pars)
        return - self.robot_calculate_score(grads)
    def __robot_iter(self, pars, grads, iterations, score):
        self.robot.iter = iterations
        self.robot.progress = self.robot.iter * 100.0 / self.robot.maxiter
        gobject.idle_add(self.update_score, -score)
        gobject.idle_add(self.OnIteration, self.robot.iter, score, grads)
        self.robot_on_iteration(pars, grads, iterations, score)
        mtree, row = self.make_trial(model_prep=self.robot.model_prep, data_prep=self.data_prep, fields={"Iteration": iterations})
        self.iter_trials.append((mtree, row))
    def robot_on_iteration(self, pars, grads, iterations, score):
        """Responds to a completed iteration; override this function to send model params etc back to gui."""
        pass
        

class ModelPrep(Anon):
    """Represents a model ready for computation with QubX.Modeling.Utils."""
    def __init__(self, model=None, modeltree=None): # one or the other, not both; modeltree is more efficient
        Anon.__init__(self, source=model, modeltree=modeltree or model.as_tree(), model=QubModel())
        self.model.from_tree(self.modeltree)
        self.K0, self.K1, self.K2, L, V, P = ModelToRateMatricesP(self.model)
    def set_rates2(self, K0, K1, K2):
        self.K0, self.K1, self.K2 = K0, K1, K2
        K012ToModel(K0, K1, K2, self.model)


class OptToolData(Requestable):
    """Caches last-used data until changed."""
    def __init__(self, get_data_prep):
        Requestable.__init__(self)
        self.get_data_prep = get_data_prep
        self.__ref = Reffer()
        self.__use_group = None
        self.__data = None
        QubX = qubx.global_namespace.QubX
        QubX.DataSource.OnChangeIdealization += self.__ref(self.__onChangeIdl)
        QubX.Data.table.OnInsert += self.__ref(self.__onChangeFileList)
        QubX.Data.table.OnRemoved += self.__ref(self.__onChangeFileList)
        QubX.Data.table.OnSet += self.__ref(self.__onSetFileList)
        QubX.Data.OnSwitch += self.__ref(self.__onSwitchDataFile)
        self.__datafile = None
        self.__onSwitchDataFile(QubX.Data, QubX.Data.file)
        QubX.Modeling.Stimulus.OnInsert += self.__ref(self.__onChangeStimulus)
        QubX.Modeling.Stimulus.OnChange += self.__ref(self.__onChangeStimulus)
        QubX.Modeling.Stimulus.OnRemoving += self.__ref(self.__onChangeStimulus)
        self.invalidate()
    def set_use_group(self, x):
        if self.__use_group != x:
            self.__use_group = x
            self.invalidate()
    use_group = property(lambda self: self.__use_group, lambda self, x: self.set_use_group(x))
    def __onChangeIdl(self, *args):
        self.invalidate()
    def __onChangeFileList(self, *args):
        if self.__use_group != None:
            self.invalidate()
    def __onSetFileList(self, i, field_name, val, prev, undoing):
        if (self.__use_group != None) and (field_name == 'Group'):
            self.invalidate()
    def __onChangeStimulus(self, *args):
        self.invalidate()
    def __onSwitchDataFile(self, Data, file):
        if file == self.__datafile: return
        if self.__datafile:
            self.__datafile.constants.OnSet -= self.__ref(self.__onSetConstant)
        self.__datafile = file
        if self.__datafile:
            self.__datafile.constants.OnSet += self.__ref(self.__onSetConstant)
    def __onSetConstant(self, r, field, val, prev, undoing):
        if self.__datafile.constants[r,'Name'] == 'Tdead (ms)':
            self.invalidate()
    def get(self, receiver):
        QubX = qubx.global_namespace.QubX
        if self.__use_group != None:
            view_segm = QubX.DataSource.get_segmentation_files(group=self.__use_group)
        else:
            view_segm = [(QubX.Data.view, QubX.DataSource.get_segmentation())]
        segs = []
        for view, segm in view_segm:
            segs.extend(segm)
        self.get_data_prep(segs, self.model_prep, wait=False, receiver=receiver)
        

class AboutPluginExample(AboutPlugin):
    __explore_featured = ['lblModelStatus', 'modelTracker']
    def __init__(self, name='Example', global_name='QubX.About.Example'):
        super(AboutPluginExample, self).__init__(name, global_name=global_name)

        # qubx uses WeakEvent for synchronous notification of listeners.
        # WeakEvent uses weak references, so it's not a big deal if a listener never unsubscribes,
        # but this means the references must be held somewhere else;
        # it requires special handling for lambda expressions, as well as bound methods,
        # which are ordinarily created and destroyed as needed.  Thus the following weak
        # reference becomes invalid by the end of the line:
        #   QubX.OnSomeEvent += self.__onSomeEvent
        # We keep a permanent reference to event handlers in a Reffer object.
        #   self.__ref(x) == x
        # and the act of calling self.__ref(x) adds it to the Reffer's internal dict (as a strong reference).
        # Thus the idiom:
        #   QubX.OnSomeEvent += self.__ref(self.__onSomeEvent)
        self.__ref = Reffer()
        
        # add components to self (it is a subclass of gtk.VBox):
        self.lblModelStatus = pack_label('', self, expand=True)
        self.lblDataStatus = pack_label('', self, expand=True)
        # they will show updated info about data and model, as they change:
        self.modelTracker = ModelTracker() # defined below
        self.modelTracker.OnSwitch += self.__ref(self.__onSwitchModel)
        self.modelTracker.OnChange += self.__ref(self.__onChangeModel)
        self.modelTracker.OnSetRate += self.__ref(self.__onSetModelRate)
        self.__onChangeModel(self.modelTracker.model) # initialize
        # QubX is the main window.  DataSource represents the active portion of data
        # (frontmost file, screen or whole file, which signal)
        QubX = qubx.pyenv.env.globals['QubX']
        self.DataSource = QubX.DataSource
        QubX.DataSource.OnChangeSamples += self.__ref(self.__onChangeSamples)
        QubX.DataSource.OnChangeIdealization += self.__ref(self.__onChangeIdealization)
        self.__onChangeData() # initialize
    def __onSwitchModel(self, models, model):
        self.__onChangeModel(model)
    def __onSetModelRate(self, r, field, val, prev, undoing):
        self.__onChangeModel(self.__model)
    def __onChangeModel(self, model):
        self.lblModelStatus.set_text("%d states" % model.states.size)
    def __onChangeSamples(self):
        self.__onChangeData()
    def __onChangeIdealization(self):
        self.__onChangeData()
    def __onChangeData(self):
        segs = self.DataSource.get_segmentation()
        nseg = len(segs)
        npoint = 0
        nideal = 0
        for seg in segs:
            for chunk in seg.chunks:
                if chunk.included:
                    npoint += chunk.n
                    ff, ll, cc = chunk.get_idealization()
                    nideal += len(ff)
        self.lblDataStatus.set_text("%d points in %d segments\n(%d events)" % (npoint, nseg, nideal))


class ToolPluginExample(ToolPlugin):
    def __init__(self, name='Test', global_name='QubX.Modeling.Test'):
        super(ToolPluginExample, self).__init__(name, global_name=global_name)
        # unused: self.__ref = Reffer()
        
        self.lblMsg = pack_label('This plugin finds the max and min value in the Data Source.', self, expand=True)
        self.panButtons = pack_item(gtk.HBox(), self)
        self.btnRun = pack_button('Run', self.panButtons, at_end=True, on_click=self.__onClickRun)
    def on_robot_begin(self):
        gobject.idle_add(self.btnRun.set_sensitive, False)
    def on_robot_end(self):
        gobject.idle_add(self.btnRun.set_sensitive, True)
    def __onClickRun(self, btn):
        # calls self.__robot_compute() on the robot (worker) thread:
        self.robot.do(self.__robot_compute)
    def __robot_compute(self):
        QubX = qubx.pyenv.env.globals['QubX']
        with self.robot.main_hold: # pause the main thread to read/write volatile info:
            units = QubX.Data.file.signals.get(QubX.DataSource.signal, 'Units')
        lo = 1e20
        hi = -lo
        segments = QubX.DataSource.get_segmentation()
        for s, seg in enumerate(segments):
            # A segment is divided in chunks, alternating included/excluded.
            # The chunks don't have sampled data yet.
            # gen_samples yields sampled chunks, and also subdivides any chunks bigger than maxlen.
            # It also takes care of pausing the main thread to read data.
            for chunk in QubX.DataSource.gen_samples(seg.chunks, self.robot.main_hold, maxlen=(1<<20)):
                if chunk.included:
                    lo = min(lo, min(chunk.samples))
                    hi = max(hi, max(chunk.samples))
            self.robot.progress = s * 100.0 / len(segments)
        # for short chunks, gen_samples is equivalent to:
        #    for chunk in seg.chunks:
        #        if chunk.included:
        #            with self.robot.main_hold:
        #                datachunk = chunk.get_samples()
        #            lo = min(lo, min(datachunk.samples))
        #            hi = max(hi, max(datachunk.samples))

        # gobject.idle_add(a, b, c, ...) calls a(b, c, ...) on the main thread, sometime later.
        # It's good for displaying intermediate and final results, messages, etc. which must
        # be done on the main thread for Win32 compatibility.
        gobject.idle_add(self.__show_results, 'Lo: %.6g %s\nHi: %.6g %s' % (lo, units, hi, units))
    def __show_results(self, results):
        self.lblMsg.set_text(results)
        

class OptToolPluginExample(OptToolPlugin):
    """Sample L{OptToolPlugin} which optimizes rate constants given idealized single-molecule data.
The algorithm is a simplification of MIL (Qin et al).
"""
    __explore_featured = ['chkPeq']
    def __init__(self, label='Example', global_name='QubX.Modeling.Example'):
        super(OptToolPluginExample, self).__init__(label, global_name=global_name, have_grads=False) # it will auto-numerical-differentiate when required
        self.__ref = Reffer()

        # add a checkbox: equilibrium or fixed entry probability
        self.chkPeq = pack_check('start at equilibrium (else use States:P0)', self.top_panel, active=True)
        
        self.append_output('This plugin demonstrates log-likelihood optimization of rate constants from idealized data, in just 85 lines.\nThe algorithm is a simplification of MIL (Qin et al, 1996).\n')

    def robot_get_pars(self):
        # this plugin doesn't handle stimulus-dependent rates, just k0
        model = self.robot.model_prep.model
        for i in xrange(model.rates.size):
            if (model.rates.get(i, 'k1') or model.rates.get(i, 'k2')
                or model.rates.get(i, 'Ligand')
                or model.rates.get(i, 'Voltage') or model.rates.get(i, 'Pressure')):
                gobject.idle_add(self.append_output, "Warning: Ignoring Ligand- and/or Voltage-dependence.\n")
        # the params are log(k0), to constrain k0 > 0
        return [log(model.rates.get(i, 'k0')) for i in xrange(model.rates.size)]
    def robot_set_pars(self, pars):
        model = self.robot.model_prep.model
        k0 = [exp(pars[i]) for i in xrange(model.rates.size)]
        # update rates in the worker-thread's model:
        for i in xrange(model.rates.size):
            model.rates.set(i, 'k0', k0[i])
        # and in the visible one, on the main thread:
        gobject.idle_add(self.__update_gui_rates, k0)
    def __update_gui_rates(self, k0):
        if self.robot.model_prep.source:
            model = self.robot.model_prep.source
            for i in xrange(model.rates.size):
                model.rates.set(i, 'k0', k0[i])

    def request_data_prep(self, receiver, segs=None, model_prep=None, last_data_prep=None):
        # this one runs in the gobject thread, as needed (cached by OptToolPlugin)
        idl = []
        nevent = 0
        dt = segs[0].sampling if segs else 1e-4
        for seg in segs:
            # first, last, class, duration -- we need only class and duration arrays
            # i.e. a segment is a sequence of dwells (events):
            #          - in a certain class (color; collection of indistinguishable states) 
            #          - for some number of seconds (by assumption, durations are even multiples of dt)
            ff, ll, cc, dd = seg.get_idealization(mark_excluded=True, get_durations=True)
            # "excluded" and un-idealized portions will have class < 0
            dd *= 1e-3 # convert from ms to sec
            idl.append( (cc, dd) )
            if any(cc >= 0): # don't count segments with un-idealized or all-excluded data
                nevent += len(cc)
        receiver(Anon(segs=segs, model_prep=model_prep, idl=idl, nevent=nevent, dt=dt))
        
    def robot_calculate_score(self, grads=None):
        QubX = qubx.pyenv.env.globals['QubX']
        idl = self.data_prep.idl
        nevent = self.data_prep.nevent
        dt = self.data_prep.dt
        if nevent == 0:
            gobject.idle_add(self.append_output, "No idealized data available (check the data source, or idealize).\n")
            return 0.0
        # read widgets while the main thread is paused:
        with self.robot.main_hold:
            usePeq = self.chkPeq.get_active()
        # get rate matrix (Q) and state colors (class)
        model = self.robot.model_prep.model
        Q = ModelToQ(model)
        N = Q.shape[0]
        class_of_state = ModelToClazz(model)
        # get equilibrium or fixed entry prob
        if usePeq:
            P0 = QtoPe(Q)
        else:
            P0 = ModelToP0(model)

        # math subroutines for transition probability and log likelihood accumulation:

        # Prob(state b at time t+dt |given| state a at time t) = A[a,b] = e^(Q*dt)[a,b]
        A_base = expm(Q*dt)
        # except we zero out columns which are ruled out by data (wrong class [color]), in this work matrix:
        A_work = matrix(zeros(shape=(N,N), dtype='float64'))
        # cache exponentiated rate matrices A(dt*nsample), since there will be a limited number of multiples of dt
        A_cache = {}
        def get_A(dur, dest_class):
            """Returns A'^(dur/dt), where A' = e^(Q*dt), except with A'[r,c]==0.0 if c!=dest_class;
caches all answers in A_cache and returns answers from cache when possible."""
            if not ((dur, dest_class) in A_cache):
                # start with one-sample transition prob. matrix:
                A_work[:] = A_base[:]
                # if data specifies destination class, then zero out other columns
                # (once per sample, it is observed in dest_class, so prob(going anywhere else in one sample) is zero)
                if dest_class >= 0:
                    A_work[:,class_of_state != c] = 0.0
                # exponentiate to repeat by number of samples
                A_cache[(dur, dest_class)] = A_work ** int(round(dur/dt))
            return A_cache[(dur, dest_class)]

        # accumulating likelihood quickly results in underflow; following Qin et al (1996),
        # we accumulate -log(scale) at each dwell, and rescale Pt
        def Scale(Pt):
            """Returns scale=1.0/sum(Pt); multiplies each component of Pt by scale."""
            mag = sum(Pt)
            if mag == 0.0:
                raise Exception('Impossible!')
            scale = 1.0 / mag
            Pt *= scale
            return scale
    
        # main loop:
        LL = 0.0
        for classes, durations in idl: # each segment
            for i, c, d in izip(count(), classes, durations): # each dwell: index, class, duration
                # transition probability into class c:
                if i == 0: # first dwell of segment: Pt[i] = P0[i], if state i is in class c
                    Pt = matrix(zeros(shape=(1,N), dtype='float64'))
                    for i in xrange(N):
                        Pt[0,i] = (class_of_state[i] == c) and P0[i] or 0.0
                else: # apply transition matrix for one sample duration (rough approximation)
                    transition_dur = min(dt, durations[i-1])
                    Pt *= get_A(transition_dur, c)
                # transition probability for dwell duration (minus transition of dt):
                dwell_dur = d - dt
                if dwell_dur > 0.0:
                    Pt *= get_A(dwell_dur, c)
                # rescale Pt toward 1.0, and accumulate log likelihood:
                LL -= log(Scale(Pt))
        return LL
        
        

class ModelTracker(object):
    """Follows the frontmost L{qubx.model.QubModel} and L{qubx.modelGTK.ModelView},
with events when anything changes.

    @ivar OnSwitch: L{WeakEvent}(QubX.Models, model) when a new file is brought to front
    @ivar OnChange: L{WeakEvent}(model) when something in the model changes (except a rate constant)
    @ivar OnSetRate: L{WeakEvent}(rate_index, field_name, value, prev, is_undoing)
"""
    def __init__(self):
        self.__ref = Reffer()
        QubX = qubx.pyenv.env.globals['QubX']
        self.Models = QubX.Models
        self.Models.OnSwitch += self.__ref(self.__onSwitchModel)
        self.__model = self.__view = None
        self.OnSwitch = WeakEvent()
        self.OnChange = WeakEvent()
        self.OnSetRate = WeakEvent()
        self.__onSwitchModel(self.Models, self.Models.file)
    def __onSwitchModel(self, models, model):
        if model == self.__model: return
        if self.__model:
            for event in (self.__model.OnSetChannelCount,
                          self.__model.OnSetIeqFv,
                          self.__model.OnSetVRev,
                          self.__model.states.OnSet,
                          self.__model.states.OnInsert,
                          self.__model.states.OnRemoved,
                          self.__model.classes.OnSet,
                          self.__model.rates.OnInsert,
                          self.__model.rates.OnRemoved,
                          self.__model.OnChangeBalanceLoops,
                          self.__model.constraints_kin.OnEdit):
                event -= self.__ref(self.__onChangeModel)
            self.__model.rates.OnSet -= self.__ref(self.__onSetRate)
        self.__model = model
        self.__view = models.view
        if self.__model:
            for event in (self.__model.OnSetChannelCount,
                          self.__model.OnSetIeqFv,
                          self.__model.OnSetVRev,
                          self.__model.states.OnSet,
                          self.__model.states.OnInsert,
                          self.__model.states.OnRemoved,
                          self.__model.classes.OnSet,
                          self.__model.rates.OnInsert,
                          self.__model.rates.OnRemoved,
                          self.__model.OnChangeBalanceLoops,
                          self.__model.constraints_kin.OnEdit):
                event += self.__ref(self.__onChangeModel)
            self.__model.rates.OnSet += self.__ref(self.__onSetRate)
        self.OnSwitch(model)
    model = property(lambda self: self.__model)
    view = property(lambda self: self.__view)
    def __onChangeModel(self, *args):
        self.OnChange(self.__model)
    def __onSetRate(self, r, field, val, prev, undoing):
        self.OnSetRate(r, field, val, prev, undoing)

        
