"""Provides a visual panel (L{qubx.faces.Face}) to optimize rate constants from ensemble measurement.

/* Copyright 2008-2013 Research Foundation State University of New York   */

/* This file is part of QUB Express.                                      */

/* QUB Express is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Express is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Express program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

"""

import cStringIO
import itertools
import gc
import operator
import os
import sys
import time
import traceback

from qubx.data_types import *
from qubx.util_types import *
from qubx.faces import *
from qubx.GTK import *
import qubx.fast.data
import qubx.fast.model
import qubx.notebook
import qubx.optimize
import qubx.optimizeGTK
import qubx.pyenv
import qubx.tree
from qubx.model import *
from qubx.task import *
from qubx.settings import Propertied, Property
from qubx.trialset import add_trial_rates, add_trial_amps, add_trial_nChannel
import macro_r_lib
from . import qubx_plugin

import scipy.optimize
from numpy import *

import ctypes
from ctypes import c_int, c_float, c_double, c_void_p, c_char_p, Structure, POINTER, CFUNCTYPE, byref, sizeof, py_object, pointer, cast
pybuf = ctypes.pythonapi.PyBuffer_FromReadWriteMemory
pybuf.restype = py_object


MIN_SECONDS_BETWEEN_FIT = 3
MAXITER = 300

class DataPrep(Anon):
    """Represents data ready for computation."""
    def __init__(self, segs, datas):
        Anon.__init__(self, segs=segs, datas=datas)



#class MacData(Requestable):
#    """Tracks data source, prepares data and stimulus as needed for LL, opt.  Data is
#[L{macro_r_lib.Data}], each item decorated with additional fields:
#    - iVoltage: index of voltage signal
#    - signal_names: list of strings
#    - file: L{qubx.data_types.QubData}
#    - ff_ll_counts: [(firsts, lasts, count)] -- resampled interval bounds, per data chunk
#
#    @ivar resample: True if data should be resampled by delta next time
#    @ivar resample_std: the "delta" of resample-by-delta
#    @ivar use_group: a Group index in the Data table, to fit multi files, or None to use the Data Source
#"""

@Propertied(Property('use_Peq', True, "True to start each segment from equilibrium; False to use States:Pr"),
            Property('use_group', False, "True to fit a group of data files", in_tree=False),
            Property('use_group_index', 1, "Group of interest in Data table"),
            Property('resample', False, "True to resample by delta y"),
            Property('resample_delta', 0.5, 'an interval with std < delta is resampled as one data point'),
            Property('Gleak', 0.0, 'Leakage conductance (pS)'),
            Property('Vleak', 0.0, 'Leak voltage (mV)'),
            Property('opt_Nchannel', False, "True to optimize the model's channel count"),
            Property('opt_cond', False, "True to optimize the model's Amps/Conds"),
            Property('opt_var', False, "True to optimize the model's Stds/CondStds"),
            Property('recursive_fit', False, "Toggle non-recursive (MacRates) and recursive (MacroR) fits"),
            # not persistent:
            Property('live', False, "True to redo fit curve when model changes", in_tree=False),
            Property('use_GPU', False, "True to use OpenCL-capable graphics card, if avail", in_tree=False)
            )
class RatesFace(qubx_plugin.OptToolPlugin):
    """Calculates log-likelihood and optimizes model parameters, a la Moffatt 2007.

@ivar OnIterationMR: L{WeakEvent}(iteration, LL, grads, K0, K1, amp, std, amp0, std0, Nchannel)
@ivar OnIteration2MR: L{WeakEvent}(iteration, LL, grads, K0, K1, K2, amp, std, amp0, std0, Nchannel)
@ivar OnLiveLL: L{WeakEvent}(LL)
"""
    __explore_featured = ['accel', 'rates', 'OnIterationMR', 'LL', 'last_LL', 'grads', 'OnLiveLL', 'fitbot', 'outStatus',
                          'get_ndata', 'add_trial_fields', 'wait_for_fits']
    def __init__(self, name='MacroR', global_name='QubX.Modeling.MacroR'):
        super(RatesFace, self).__init__(name, global_name,
                                        "qubx_macror_optimize",
                                        {"max step" : 0.1,
                                         "num grad dx" : 1e-4,
                                         "conv ll" : 1e-3,
                                         "conv grad" : 1e-3})
        self.__ref = Reffer()
        self.__invalid_score = False
        self.__invalid_live = True
        self.accel = 1 # (0: single; 1: openmp; 2: opencl)
        self.QubX = qubx.pyenv.env.globals['QubX']
        self.QubX.OnQuit += self.__ref(self.mac_finalize)
        self.propertied_connect_settings('MacroR')
        self.data.use_group = self.use_group_index if self.use_group else None
        self.rates = Product(self.__get_rates) # overriding OptToolPlugin.rates
        self.rates.OnInvalidate += self.__ref(self.__onInvalidateRates)
        self.OnIterationMR = WeakEvent()
        self.OnIteration2MR = WeakEvent()

        self.robot.working = False
        self.robot.workspaces = None
        self.robot.optNchannel = self.robot.optCond = self.robot.optVar = False
        self.iterations = 0
        self.LL = self.last_LL = 0.0
        self.grads = []
        self.__Ndata = 0
        self.OnLiveLL = WeakEvent()

        # When calculation is fast enough, showing fit curves can be the slowest part of an iteration.
        # Ideally we'd skip displaying an iteration's curves if the optimizer has already finished the next iteration.
        # Showing curves on a separate thread makes it easier:
        self.fitbot = Robot('Apply Fit') # not in task list
        self.fitbot.OnException += self.__ref(self.on_exception)
        self.fitbot.serial = self.fitbot.last_serial = 0
        # We also skip iterations when too few seconds have elapsed:
        self.__last_fits_time = datetime.datetime.now()
        # rewind the clock so fit curves can show right away (there must be a simpler way):
        if self.__last_fits_time.hour:
            self.__last_fits_time = self.__last_fits_time.replace(hour=self.__last_fits_time.hour-1)
        # else: it will not show any fits for the first few seconds oh well

        # these figures will be published through Admin:Notebook
        qubx.notebook.Notebook.register_auto('MacroR_eval.Messages', 'Messages, on MacroR: Get LL', False)
        qubx.notebook.Notebook.register_auto('MacroR_eval.Model.Picture', 'Model picture, on MacroR: Get LL', False)
        qubx.notebook.Notebook.register_auto('MacroR_eval.Model.PictureNO', 'Model picture, no overlays, on MacroR: Get LL', False)
        qubx.notebook.Notebook.register_auto('MacroR_eval.Data.Picture', 'Data picture, on MacroR: Get LL', False)
        qubx.notebook.Notebook.register_auto('MacroR_eval.Model.Classes', 'Classes table, on MacroR: Get LL', False)
        qubx.notebook.Notebook.register_auto('MacroR_eval.Model.Rates', 'Rates table, on MacroR: Get LL', False)
        
        qubx.notebook.Notebook.register_auto('MacroR_opt.Iterations', 'Iteration log, on MacroR: Optimize', True)
        qubx.notebook.Notebook.register_auto('MacroR_opt.Messages', 'Messages, on MacroR: Optimize', True)
        qubx.notebook.Notebook.register_auto('MacroR_opt.Model.Picture', 'Model picture, on MacroR: Optimize', True)
        qubx.notebook.Notebook.register_auto('MacroR_opt.Model.PictureNO', 'Model picture, no overlays, on MacroR: Optimize', False)
        qubx.notebook.Notebook.register_auto('MacroR_opt.Data.Picture', 'Data picture, on MacroR: Optimize', True)
        qubx.notebook.Notebook.register_auto('MacroR_opt.Model.Classes', 'Classes table, on MacroR: Optimize', True)
        qubx.notebook.Notebook.register_auto('MacroR_opt.Model.Rates', 'Rates table, on MacroR: Optimize', False)
        qubx.notebook.Notebook.register_auto('MacroR_opt.Model.Constraints', 'Kinetic Constraints table, on MacroR: Optimize', True)

        col = pack_item(gtk.VBox(), self.top_panel)
        self.chkPeq = pack_check('start at equilibrium p', col)
        self.chkPeq.set_tooltip_text('otherwise use States:Pr')
        self.propertied_connect_check('use_Peq', self.chkPeq)
        h = pack_item(gtk.HBox(), col)
        self.chkResample = pack_check('resample by delta:', h)
        self.chkResample.set_tooltip_text('skip slow changes:\ncompress data with std.dev < delta into a single point')
        self.propertied_connect_check('resample', self.chkResample)
        self.txtResample = pack_item(NumEntry(self.resample_delta, acceptFloatGreaterThan(0.0), '%.3g', width_chars=6), h)
        self.propertied_connect_NumEntry('resample_delta', self.txtResample)
        
        pack_space(self.top_panel, x=20)
        col = pack_item(gtk.VBox(), self.top_panel)
        h = pack_item(gtk.HBox(), col)
        pack_label('Gleak:', h)
        self.txtGleak = pack_item(NumEntry(self.Gleak, acceptFloat, '%.3g', width_chars=6), h, at_end=True)
        self.propertied_connect_NumEntry('Gleak', self.txtGleak)
        h = pack_item(gtk.HBox(), col)
        pack_label('Vleak:', h)
        self.txtVleak = pack_item(NumEntry(self.Vleak, acceptFloat, '%.3g', width_chars=6), h, at_end=True)
        self.propertied_connect_NumEntry('Vleak', self.txtVleak)

        pack_space(self.top_panel, expand=True)
        col = pack_item(gtk.VBox(True), self.top_panel)
        lbl_d = pack_label('Multi-data:', col)
        #lbl_m = pack_label('Multi-model:', col)
        col = pack_item(gtk.VBox(True), self.top_panel)
        h = pack_item(gtk.HBox(), col)
        self.chkUseGroup = pack_check('Group', h)
        self.propertied_connect_check('use_group', self.chkUseGroup)
        self.txtUseGroup = pack_item(qubx.GTK.NumEntry(self.use_group_index, acceptIntGreaterThan(-1), width_chars=3), h)
        self.txtUseGroup.set_sensitive(self.use_group)
        self.propertied_connect_NumEntry('use_group_index', self.txtUseGroup)
        for widget in (lbl_d, self.chkUseGroup, self.txtUseGroup):
            widget.set_tooltip_text('To select data files, set their Group number in the Data table.\nThey will be fit to common parameters')
        #h = pack_item(gtk.HBox(), col)
        #self.chkModelsGroup = pack_check('Group', h)
        #self.propertied_connect_check('multi_model', self.chkModelsGroup)
        #self.txtModelsGroup = pack_item(qubx.GTK.NumEntry(self.multi_model_group, acceptIntGreaterThan(-1), width_chars=3), h)
        #self.txtModelsGroup.set_sensitive(self.multi_model)
        #self.propertied_connect_NumEntry('multi_model_group', self.txtModelsGroup)
        #for widget in (lbl_m, self.chkModelsGroup, self.txtModelsGroup):
        #    widget.set_tooltip_text("To select models, set their Group number in the Models table.\nData will be fit to the sum of all models' currents.")
        
        pack_space(self.top_panel, expand=True)
        col = pack_item(gtk.VBox(True), self.top_panel)
        self.chkLive = pack_check('live   ', col)
        self.chkLive.set_tooltip_text('update LL and fit curve when the model changes')
        self.propertied_connect_check('live', self.chkLive)
        self.chkRecursiveFit = pack_check('rec. fit', col)
        self.chkRecursiveFit.set_tooltip_text('on: shows the recursive prior distribution;\noff: shows the MacRates (non-recursive) fit')
        self.propertied_connect_check('recursive_fit', self.chkRecursiveFit)
        self.chkGPU = pack_check('use GPU   ', col, show=False)
        self.chkGPU.set_tooltip_text('use graphics card if available; faster but less accurate')
        self.propertied_connect_check('use_GPU', self.chkGPU)
        
        self.lblScore.set_text('LL:')
        self.btnGetScore.set_label('Get LL')

        pack_label('Optimize:', self.right_panel)
        self.chkOptNchannel = pack_check('channel count', self.right_panel)
        self.propertied_connect_check('opt_Nchannel', self.chkOptNchannel)
        self.chkOptCond = pack_check('conductance', self.right_panel)
        self.propertied_connect_check('opt_cond', self.chkOptCond)
        self.chkOptVar = pack_check('variance', self.right_panel)
        self.propertied_connect_check('opt_var', self.chkOptVar)
        self.chkOptRates = pack_check('rates', self.right_panel, active=True)
        self.chkOptRates.set_tooltip_text('To hold a rate constant, right-click it, "Add Constraint...," "Fix..."')
        self.chkOptRates.set_sensitive(False)
        
        self.outStatus = TextViewAppender(self.txtStatus)
        self.outStatus.write('Based on "Estimation of ion channel kinetics from fluctuations of macroscopic currents," Moffatt L., Biophys J. 2007 Jul 1; 93(1):74-91\n')
        self.outStatus.OnWrite += self.__ref(self.__onWriteStatus)
        self.__nb_iter_buf = None
    def mac_finalize(self):
        """Stops the worker threads; call this in qubx_plugin_fini."""
        self.robot.stop()
        self.fitbot.stop()
        self.data = self.data_prep = None
    def on_exception(self, robot, typ, val, trace):
        traceback.print_exception(typ, val, trace)
        traceback.print_exception(typ, val, trace, file=self.outStatus)
    def __onWriteStatus(self, msg):
        if not (self.__nb_iter_buf is None):
            self.__nb_iter_buf.write(msg)
    def propertied_set(self, value, name):
        if (name, value) == ('use_group_index', None):
            self.use_group = False
            return
        super(RatesFace, self).propertied_set(value, name)
        if name in ('use_Peq', 'Gleak', 'Vleak'):
            self.invalidate_controls()
        if name in ('resample', 'resample_delta'):
            self.data.invalidate()
        if name in ('use_group', 'use_group_index'):
            self.data.use_group = self.use_group_index if self.use_group else None
            self.txtUseGroup.set_sensitive(self.use_group)
        if (name == 'use_group_index') and not self.use_group:
            self.use_group = True
        if name == 'live':
            if value:
                self.invalidate_controls()
        if name == 'use_GPU':
            self.invalidate_controls()
    def __onInvalidateRates(self, prod):
        self.invalidate_controls()
    def invalidate_controls(self, gray=True):
        """Changes LL text color gray (outdated) or black (up-to-date), and starts computation if showing in live mode."""
        if self.__invalid_score != gray:
            self.__invalid_score = gray
            color = gdk.color_parse(gray and "#999999" or "#000000")
            self.txtScore.modify_text(gtk.STATE_NORMAL, color)
        if gray and self.showing and self.chkLive.get_active():
            self.__queue_robot_live()
    def __queue_robot_live(self):
        if self.recursive_fit:
            self.__invalid_live = True
            if not (self.robot.working or self.optimizing):
                self.__capture_robot_settings()
                self.robot.do(self.robot_live)
        elif hasattr(qubx.global_namespace.QubX.Modeling, 'MacRates'):
            MacRates = qubx.global_namespace.QubX.Modeling.MacRates
            if not MacRates.live:
                MacRates.live = True
                MacRates.live = False
    def __sensitize(self, sens):
        try:
            self.chkPeq.set_sensitive(sens)
        except:
            pass
    def on_robot_begin(self):
        super(RatesFace, self).on_robot_begin()
        gobject.idle_add(self.__sensitize, False)
    def on_robot_end(self):
        super(RatesFace, self).on_robot_end()
        gobject.idle_add(self.__sensitize, True)
    def fitbot_send_datapic(self, auto_id):
        gobject.idle_add(lambda: qubx.notebook.Notebook.send(self.QubX.Data.view.nbPicture, auto_id=auto_id))
    def get_ndata(self, data_prep=None):
        if data_prep and data_prep.workspaces:
            return reduce(operator.add, (ws.data.Ndata for ws in data_prep.workspaces), 0)
        else:
            return self.__Ndata
    def __get_rates(self):
        model_prep = self.model_prep.val
        signal_names = iunion((prep.source or prep.model).variables for prep in model_prep.preps)
        return [ModelToRateMatricesP(prep.source or prep.model, signal_names)[:3] for prep in model_prep.preps]
    def add_trial_fields(self, row, mtree, prefix='', model_prep=None, data_prep=None):
        add_trial_rates(row, mtree, prefix)
        if (not prefix) or self.robot.optNchannel:
            add_trial_nChannel(row, mtree, prefix)
        if (not prefix) or self.robot.optCond or self.robot.optVar:
            add_trial_amps(row, mtree, prefix)
    def __onGotLL(self, LL):
        self.LL = LL
        if not self.optimizing:
            qubx.notebook.Notebook.send(qubx.notebook.NbText("""MacroR: Get LL: %.6g""" % LL), auto_id='MacroR_eval.Messages')
            if self.QubX.Models.file in self.robot.model_prep.sources: ### TODO: send all sources
                qubx.notebook.Notebook.send(self.QubX.Models.view.nbPicture, auto_id='MacroR_eval.Model.Picture')
                qubx.notebook.Notebook.send(self.QubX.Models.view.nbPictureNO, auto_id='MacroR_eval.Model.PictureNO')
                qubx.notebook.Notebook.send(self.QubX.Tables.find_view('Classes').nbTable, auto_id='MacroR_eval.Model.Classes')
                qubx.notebook.Notebook.send(self.QubX.Tables.find_view('Rates').nbTable, auto_id='MacroR_eval.Model.Rates')
            # fit robot send datapic with updated fit curves
            fitbot = qubx.global_namespace.QubX.Modeling.MacRates.fitbot if ((not self.recursive_fit) and hasattr(qubx.global_namespace.QubX.Modeling, 'MacRates')) else self.fitbot
            fitbot.do(self.fitbot_send_datapic, 'MacroR_eval.Data.Picture')
            if self.__invalid_live:
                self.__queue_robot_live()
            else:
                self.invalidate_controls(False)
            self.set_score(LL)
            self.last_LL = LL
    def calculate_score(self, model_prep=None, data_prep=None, wait=True, receiver=lambda score: None, on_trial=None, **kw):
        self.__capture_robot_settings()
        return super(RatesFace, self).calculate_score(model_prep, data_prep, wait, receiver, on_trial, **kw)
    def optimize(self, model_prep=None, data_prep=None, wait=True, receiver=None, ext_return=False, on_trial=None):
        self.__nb_iter_buf = cStringIO.StringIO()
        self.__capture_robot_settings()
        self.QubX.Models.view.optimizing = True
        return super(RatesFace, self).optimize(model_prep, data_prep, wait, receiver, ext_return, on_trial)
    def __onOptDone(self, LL, iterations, grads, Hessian, workspaces, model_prep, data_prep):
        self.QubX.Models.view.optimizing = False
        self.LL = LL
        self.iterations = iterations
        self.grads = grads
        iteration_messages = self.__nb_iter_buf.getvalue() if self.__nb_iter_buf else ""
        self.__nb_iter_buf = cStringIO.StringIO() # to capture final output for notebook
        for source in model_prep.sources:
            source.undoStack.seal_undo('set rates')
        Ndata = self.__Ndata
        self.append_output('===================================================\n')
        self.outStatus.write("""MacroR: finished optimizing %d points after %d iterations.
Data:           %s
Model:          %s
P0:             %s
Resample:       %s
Gleak:          %.4g
Vleak:          %.4g
GPU:            %s
Optimize:       %s
Gradient:       %s
""" % (Ndata,
                      iterations,
                      '\n           '.join(os.path.split(ws.data.file.path)[1] for ws in workspaces),
                      model_prep.sources and ', '.join([os.path.split(source.path)[1] for source in model_prep.sources]) or "from script",
                      self.robot.Peq and "equilibrium" or "fixed (States table)",
                      self.data.resample and ("by delta = %.4g" % self.data.resample_std) or "no",
                      self.robot.Gleak,
                      self.robot.Vleak,
                      self.chkGPU.get_active() and "yes" or "no",
                      ((self.robot.optNchannel and "channel count, " or "") +
                       (self.robot.optCond and "conductance, " or "") +
                       (self.robot.optVar and "variance, " or "") +
                       "rates"),
                      ', '.join('%8.5g' % x for x in grads)))
        if workspaces:
            self.__set_model_confidence(workspaces[0], model_prep)
        self.outStatus.write('\nLL: ')
        self.outStatus.write_bold('%.5g' % LL)
        if self.last_LL:
            self.outStatus.write('        last LL: %.5g' % self.last_LL)
        self.append_output('\n===================================================\n')
        self.invalidate_controls(False)
        qubx.notebook.Notebook.send(qubx.notebook.NbText(iteration_messages), auto_id='MacroR_opt.Iterations')
        qubx.notebook.Notebook.send(qubx.notebook.NbText(self.__nb_iter_buf.getvalue()), auto_id='MacroR_opt.Messages')
        self.__nb_iter_buf = None
        if self.QubX.Models.file in model_prep.sources: ### TODO: send all source models
            qubx.notebook.Notebook.send(self.QubX.Models.view.nbPicture, auto_id='MacroR_opt.Model.Picture')
            qubx.notebook.Notebook.send(self.QubX.Models.view.nbPictureNO, auto_id='MacroR_opt.Model.PictureNO')
            qubx.notebook.Notebook.send(self.QubX.Tables.find_view('Classes').nbTable, auto_id='MacroR_opt.Model.Classes')
            qubx.notebook.Notebook.send(self.QubX.Tables.find_view('Rates').nbTable, auto_id='MacroR_opt.Model.Rates')
            qubx.notebook.Notebook.send(self.QubX.Tables.find_view('Kinetic Constraints').nbTable, auto_id='MacroR_opt.Model.Constraints')
        # fit robot send datapic with updated fit curves
        fitbot = qubx.global_namespace.QubX.Modeling.MacRates.fitbot if ((not self.recursive_fit) and hasattr(qubx.global_namespace.QubX.Modeling, 'MacRates')) else self.fitbot
        fitbot.do(self.fitbot_send_datapic, 'MacroR_opt.Data.Picture')
        self.last_opt_model = workspaces[0].param.modelinfo[0].fastmodel if workspaces else None
        self.last_LL = LL
    def __set_model_confidence(self, ws, model_prep):
        for mi in ws.param.modelinfo:
            self.outStatus.write("\n", scroll=False) # model name?
            if self.robot.optNchannel:
                self.outStatus.write("Channel count:  ", scroll=False)
                self.outStatus.write_bold('%i' % mi.fastmodel.Nchannel, scroll=False)
                self.outStatus.write(" +/- %.1f\n" % mi.fastmodel.Nchannel_err, scroll=False)
            if self.robot.optCond or self.robot.optVar:
                amp_lbl, std_lbl = ('Amp', 'Variance') if (not mi.fastmodel.IeqFv) else ('Cond', 'CondVar')
                if mi.fastmodel.IeqFv and mi.fastmodel.iVoltage:
                    if self.robot.optCond:
                        self.outStatus.write('Baseline Amp:  ', scroll=False)
                        self.outStatus.write_bold('%.5g' % mi.fastmodel.amp0, scroll=False)
                        self.outStatus.write(' +/- %.5g\n' % mi.fastmodel.amp0_err, scroll=False)
                    if self.robot.optVar:
                        self.outStatus.write('Baseline Variance: ', scroll=False)
                        self.outStatus.write_bold('%.5g' % mi.fastmodel.var0, scroll=False)
                        self.outStatus.write(' +/- %.5g\n' % mi.fastmodel.var0_err, scroll=False)
                self.outStatus.write('%7s' % 'State', scroll=False)
                if self.robot.optCond:
                    self.outStatus.write_bold('%14s' % amp_lbl, scroll=False)
                    self.outStatus.write('%14s' % (amp_lbl+' err'), scroll=False)
                if self.robot.optVar:
                    self.outStatus.write_bold('%14s' % std_lbl, scroll=False)
                    self.outStatus.write('%14s' % (std_lbl+' err'), scroll=False)
                self.outStatus.write('\n', scroll=False)
                for c in xrange(mi.fastmodel.Nclass):
                    self.outStatus.write('%7d' % c)
                    if self.robot.optCond:
                        self.outStatus.write_bold('%14.5g' % mi.fastmodel.amp[c], scroll=False)
                        self.outStatus.write('%14.5g' % mi.fastmodel.amp_err[c], scroll=False)
                    if self.robot.optVar:
                        self.outStatus.write_bold('%14.5g' % mi.fastmodel.var[c], scroll=False)
                        self.outStatus.write('%14.5g' % mi.fastmodel.var[c], scroll=False)
                    self.outStatus.write('\n', scroll=False)
            self.outStatus.write('%10s%7s' % ("Rate From", "To"), scroll=False)
            self.outStatus.write_bold('%14s' % "k0", scroll=False)
            self.outStatus.write('%11s' % "+/-", scroll=False)
            self.outStatus.write_bold('%14s' % 'k1', scroll=False)
            self.outStatus.write('%11s' % "+/-", scroll=False)
            self.outStatus.write_bold('%14s' % 'k2', scroll=False)
            self.outStatus.write('%11s' % "+/-", scroll=False)
        for mi, source in itertools.izip(ws.param.modelinfo, model_prep.sources):
            self.outStatus.write('\n', scroll=False)
            for rate in source.rates:
                a,b = rate.From, rate.To
                source.rates[rate.Index, 'k0 std'] = mi.fastmodel.K0_err[a,b]
                source.rates[rate.Index, 'k1 std'] = mi.fastmodel.K1_err[a,b]
                source.rates[rate.Index, 'k2 std'] = mi.fastmodel.K2_err[a,b]
                self.outStatus.write('%10d%7d' % (a, b), scroll=False)
                self.outStatus.write_bold('%14.5g' % rate.k0, scroll=False)
                self.outStatus.write('%11.3g' % mi.fastmodel.K0_err[a,b], scroll=False)
                self.outStatus.write_bold('%14.5g' % rate.k1, scroll=False)
                self.outStatus.write('%11.3g' % mi.fastmodel.K1_err[a,b], scroll=False)
                self.outStatus.write_bold('%14.5g' % rate.k2, scroll=False)
                self.outStatus.write('%11.3g' % mi.fastmodel.K2_err[a,b], scroll=False)
                self.outStatus.write('\n')
    def __onIteration(self, iteration, LL, grads, optNchannel, workspaces, modelinfo, K0s, K1s, K2s, amps, stds, amp0s, std0s, Nchannels):
        self.LL = LL
        self.iterations = iteration
        self.grads = grads
        for mi, K0, K1, K2, amp, std, amp0, std0, Nchannel in itertools.izip(modelinfo, K0s, K1s, K2s, amps, stds, amp0s, std0s, Nchannels):
            model = mi.prep.source
            if not model:
                continue
            K012ToModel(K0, K1, K2, model)
            if model.IeqFv:
                for c,a in enumerate(amp):
                    model.classes.set(c, 'Cond', a)
                    model.classes.set(c, 'CondStd', std[c])
                model.classes.set(0, 'Amp', amp0)
                model.classes.set(0, 'Std', std0)
            else:
                for c,a in enumerate(amp):
                    model.classes.set(c, 'Amp', a)
                    model.classes.set(c, 'Std', std[c])
            model.set_channelCount(Nchannel, seal=False) # seal: don't want undo-per-iteration
            if (self.__Ndata > 1) and optNchannel and (len(modelinfo) == 1):
                for i, ws in enumerate(workspaces):
                    datafile = qubx.global_namespace.QubX.Data.views[ws.param.data.file_index].file
                    try:
                        cc_ix = datafile.constants.index('Channel Count')
                        datafile.constants[cc_ix, 'Value'] = ws.param.modelinfo[0].fastmodel.Nchannel
                    except KeyError:
                        datafile.constants.append({'Name' : 'Channel Count', 'Value' : ws.param.modelinfo[0].fastmodel.Nchannel})
        self.invalidate_controls(False)
        self.outStatus.write('Iter: %i\tLL: %8.5g\tgrads: %s\n' % (iteration, LL, ', '.join("%8.5g"%g for g in grads)))
        self.OnIterationMR(iteration, LL, grads, K0s, K1s, amps, stds, amp0s, std0s, Nchannels)
        self.OnIteration2MR(iteration, LL, grads, K0s, K1s, K2s, amps, stds, amp0s, std0s, Nchannels)
    def on_interrupt(self, task, cancel):
        cancel() # the flag is good enough, while the KeyboardInterrupt can bewilder it
        for ws in self.robot.workspaces:
            for mi in ws.param.modelinfo:
                mi.fastmodel.obj[0].stop_flag = 1
    def get_model_prep(self, model=None):
        mp = super(RatesFace, self).get_model_prep(model)
        mp.models = [mp.model]
        mp.sources = [mp.source] if mp.source else []
        mp.preps = [mp] # please please please garbage collect despite this cycle
        mp.set_rates2 = set_modelprep_rates(mp)
        return mp
    def request_data_prep(self, receiver, segs, model_prep, last_data_prep=None, resample=None, resample_std=None):
        QubX = qubx.global_namespace.QubX
        view_of_file = dict([(view.file, view) for view in QubX.Data.views])
        view_segm = []
        vs_by_file = {}
        for seg in segs:
            if not (seg.file in vs_by_file):
                vs = (view_of_file[seg.file], [])
                view_segm.append(vs)
                vs_by_file[seg.file] = vs
            vs_by_file[seg.file][1].append(seg)
        if not view_segm:
            if receiver:
                gobject.idle_add(receiver, DataPrep(segs, []))
            return

        resamp = QubX.Modeling.MacroR.resample if (resample is None) else resample
        resamp_std = QubX.Modeling.MacroR.resample_delta if (resample_std is None) else resample_std

        datas = []
        data_prep = DataPrep(segs, datas)
        self.last_LL = 0.0
        # segment is divided in chunks, alternating included/excluded (data / no data)
        # each segment, we request idl of each signal, each chunk
        # this could involve re-idealization, indefinite time, so result comes via callback
        # we provide generator instance as callback, repeatedly, to process segments in a loop
        
        # gobject thread:                               stimulus robot:
        #   rcv_stims.next()
        #      repeat:
        #         stimulus.call_with_stimulus_idl
        #         yield...
        #   [get(self, ...) returns]
        #                                                  idealize stim if needed
        #                                                  idle_add(rcv_stims.send, ...)
        #         ...process chunks
        #      yield None

        rcv_stims = None # will have instance of receive_stims()
        def rcv_hack(*args): # can send only one argument into a generator
            rcv_stims.send(args)
        def receive_stims():
            stim_names = iunion((prep.source or prep.model).variables for prep in model_prep.preps)
            for dataview, segsrc in view_segm:
                # re-use L{macro_r_lib.Data} object: (single-file only)
                if last_data_prep and len(last_data_prep.datas) and (len(view_segm) == 1):
                    data = last_data_prep.datas[0]
                    data.reset(len(stim_names))
                else:
                    data = macro_r_lib.Data(len(stim_names), file_index=QubX.Data.views.index(dataview)) # , data_cap_samples)
            
                sampling_sec = dataview.file.sampling
                if 'Voltage' in stim_names:
                    data.iVoltage = stim_names.index('Voltage') + 1
                else:
                    data.iVoltage = 0
                data.signal_names = ['Current'] + stim_names
                data.ff_ll_counts = []

                iseg = -1
                for seg in segsrc:
                    data.file = seg.file
                    data.iSignal = seg.signal
                    QubX.Modeling.Stimulus.call_with_stimulus_idl(rcv_hack, [(chunk.f, chunk.l) for chunk in seg.chunks], datafile=dataview.file, names=stim_names)
                    stim_names, stimidl = (yield)
                    chunkidl = [ Anon(dwellCounts=[], classeses=[], durationses=[], ampses=[]) for chunk in seg.chunks]
                    missing = False
                    for segs_flcda in stimidl: # for each signal
                        for ichunk, flcda in enumerate(segs_flcda): # each chunk
                            ff, ll, cc, dd, aa  = flcda
                            if not len(cc):
                                missing = True
                            chunkidl[ichunk].dwellCounts.append(len(cc))
                            chunkidl[ichunk].classeses.append(cc)
                            chunkidl[ichunk].durationses.append(dd)
                            chunkidl[ichunk].ampses.append(aa)
                    if missing:
                        continue
                
                    iseg += 1
                    data.append_segment()
                    for ichunk, chunk in enumerate(seg.chunks):
                        if chunk.included:
                            Nd = chunk.n
                            chunkoff = chunk.f - seg.f
                            xx = arange(chunkoff, chunkoff+Nd, dtype='float32')
                            xx *= sampling_sec
                            yy = zeros(shape=(Nd,), dtype='float32')
                            ii = 0
                            for ch in generate_chunk_samples([chunk]):
                                yy[ii:ii+ch.n] = ch.samples
                                ii += ch.n
                            if resamp:
                                means, stds, ff, ll, closest = qubx.fast.data.adaptive_resample(yy, resamp_std)
                                xx = array(xx[closest], dtype='float32', copy=True)
                                yy = array(yy[closest], dtype='float32', copy=True)
                                ff += chunk.f
                                ll += chunk.f
                                Nd = len(means)
                            else:
                                ff = arange(seg.f+chunkoff, seg.f+chunkoff+Nd, dtype='int32')
                                ll = ff
                            data.ff_ll_counts.append( (ff, ll, Nd) )
                        else:
                            Nd = 0
                            xx = yy = None
                        data.append_chunk(iseg, (chunk.l - chunk.f + 1) * sampling_sec, Nd, xx, yy, None,
                                          chunkidl[ichunk].dwellCounts, chunkidl[ichunk].classeses,
                                          chunkidl[ichunk].durationses, chunkidl[ichunk].ampses,
                                          (not resamp) and sampling_sec or 0.0)
                datas.append(data)
            gobject.idle_add(receiver, data_prep)
            yield None

        rcv_stims = receive_stims()
        rcv_stims.next()
    def __onGPUFail(self):
        self.chkGPU.set_active(False)
        self.chkGPU.set_sensitive(False)
        self.chkGPU.set_tooltip_text('no OpenCL-compatible graphics processor is available')
    def wait_for_fits(self):
        """Waits until all pending work is done on the thread which updates fit curves, then returns."""
        fitbot = qubx.global_namespace.QubX.Modeling.MacRates.fitbot if ((not self.recursive_fit) and hasattr(qubx.global_namespace.QubX.Modeling, 'MacRates')) else self.fitbot
        while fitbot.last_serial != fitbot.serial:
            qubx.pyenv.env.process_events()
    def __capture_robot_settings(self):
        """Should happen at the same time the modelprep is captured, e.g. at button press"""
        self.robot.Peq = self.use_Peq
        self.robot.GPU = self.chkGPU.get_active()
        self.robot.Gleak = self.Gleak
        self.robot.Vleak = self.Vleak
        self.robot.optNchannel = self.opt_Nchannel
        self.robot.optCond = self.opt_cond
        self.robot.optVar = self.opt_var
        self.data.resample = self.resample
        self.data.resample_std = self.resample_delta
        self.robot.K012 = self.rates.val
        self.robot.rates_source = qubx.global_namespace.QubX.Models.file
        if self.robot.GPU:
            self.robot.accel = self.accel and (self.accel + int(self.robot.GPU)) or 0
        else:
            self.robot.accel = self.accel
    def robot_setup(self):
        model_prep = self.robot.model_prep
        self.robot.models = model_prep.models
        self.robot.modelinfo = [Anon(model=prep.model, prep=prep) for prep in model_prep.preps if prep.model.states.size]
        self.data.model_prep = model_prep
        
        for mi in self.robot.modelinfo:
            mi.clazz = ModelToClazz(mi.model)
            mi.Nclass = int(max(mi.clazz) + 1) if mi.clazz.shape[0] else 0

        datas = self.robot.data = self.data_prep.datas # if Nclass else []
        if self.robot.workspaces:
            for ws in self.robot.workspaces:
                try:
                    ws.fastmodel.dispose() # work around callback-related memory leak
                except:
                    pass
        workspaces = self.robot.workspaces = []
        self.__Ndata = reduce(operator.add, (data.Ndata for data in datas), 0)

        for data in datas:
            stimclasses = data.get_stimclasses()
            Nstim = data.Nstim
            Nsig = data.Nsig
            sc_frac = data.get_stimclass_frac()
            param = macro_r_lib.Param()
            param.modelinfo = []
            param.data = data
            param.useVar = True ##
            param.optNchannel = self.robot.optNchannel
            param.accel = self.robot.accel

            for im, mi in enumerate(self.robot.modelinfo):
                fastmodel = ModelToFast(mi.model, data.signal_names, self.robot_on_status)
                fastmodel.usePeq = self.robot.Peq
                fastmodel.iVoltage = data.iVoltage
                fastmodel.Gleak = self.robot.Gleak
                fastmodel.Vleak = self.robot.Vleak
                K0, K1, K2, L, V, P = ModelToRateMatricesP(mi.model, data.signal_names)
                fastmodel.K0[:], fastmodel.K1[:], fastmodel.K2[:] = K0, K1, K2 = self.robot.K012[im]

                if len(datas) > 1:
                    try:
                        fastmodel.Nchannel = qubx.global_namespace.QubX.Data.views[data.file_index].file.constants.get_by_name('Channel Count', 'Value')
                    except:
                        pass # use value from model
        
                ratesm = qubx.fast.model.StimRates(fastmodel, Nstim, Nsig, stimclasses, sc_frac, 0.0) # skip ratesm.AA generation
                A, B = mi.model.constraints_kin.get_matrices_p(K0, K1, K2, L, V, P, auto=False)
                if A != None:
                    ratesm.cns.Ain[:A.shape[0],:A.shape[1]] = A
                    ratesm.cns.Bin[:B.shape[0]] = B.flatten()
                    ratesm.cns.Ncns = B.shape[0]
                else:
                    ratesm.cns.Ncns = 0
                    
                ampm = qubx.fast.model.StimAmps(fastmodel, Nstim, Nsig, stimclasses, sc_frac)
                if not self.robot.optCond:
                    for c in xrange(fastmodel.Nclass):
                        ampm.fix_amp(c)
                    if fastmodel.IeqFv:
                        ampm.fix_amp0()
                if not self.robot.optVar:
                    for c in xrange(fastmodel.Nclass):
                        ampm.fix_var(c)
                    if fastmodel.IeqFv:
                        ampm.fix_var0()
                param.add_model(fastmodel, ampm, ratesm)
                param.modelinfo.append(mi)
                mi.fastmodel = fastmodel
                mi.Nchannel = fastmodel.Nchannel

            if param.models:
                param.setup_constraints()
                workspaces.append(Anon(param=param, data=data, modelinfo=param.modelinfo, L=L, V=V))
    def robot_live(self, model_prep=None):
        if self.__invalid_live and not self.robot.working:
            if model_prep is None:
                with self.robot.main_hold:
                    mp = self.model_prep.val
            else:
                mp = model_prep
            self.robot_setup_data(mp, None)
            LL = self.robot_calculate_score()
            gobject.idle_add(self.OnLiveLL, LL)
    def robot_calculate_score(self, grads=None):
        LL = 0.0
        try:
            self.robot.working = True
            self.__invalid_live = False
            if not self.optimizing:
                self.robot_setup()
            LL = 0.0
            for ws in self.robot.workspaces:
                try:
                    LL += macro_r_lib.mac_ll(ws.param)
                except:
                    gobject.idle_add(self.outStatus.write, traceback.format_exc() + '\n')
                    LL += -1.0
                if ws.param.gpu_fail:
                    gobject.idle_add(self.__onGPUFail)
                    self.robot.accel -= 1
                    for ws2 in self.robot.workspaces:
                        ws2.param.accel = self.robot.accel
            self.robot_queue_fits(0, True)
        except:
            self.robot.send_exception()
        gobject.idle_add(self.__onGotLL, LL)
        self.robot.working = False
        return LL
    def robot_get_pars(self):
        """Returns the list of initial parameter values, on worker thread; override this method.
Derive parameter values from self.robot.model and/or from your own widgets."""
        self.robot_setup()
        if self.robot.workspaces:
            pars = [float(x) for x in self.robot.workspaces[0].param.fPars[:self.robot.workspaces[0].param.NfPar]] # get as python array
        else:
            pars = []
        # fastmodel has room for one channel count, used for first file; append a param for the rest
        if (len(self.robot.workspaces) > 1) and self.robot.optNchannel and (len(self.robot.models) == 1):
            pars.extend([1.0] * (len(self.robot.workspaces)-1))
            self.__scale_Nchannel = [log(ws.param.Nchannel) for ws in self.robot.workspaces[1:]]
        return array(pars)
    def robot_set_pars(self, pars):
        if (pars != None) and self.robot.workspaces:
            NfPar = self.robot.workspaces[0].param.NfPar
            self.robot.workspaces[0].param.fPars[:NfPar] = pars[:NfPar]
            self.robot.workspaces[0].param.do_fPars_to_model()
            if (len(self.robot.workspaces) > 1) and self.robot.optNchannel and (len(self.robot.models) == 1):
                ccpar = pars[-(len(self.robot.workspaces)-1):]
                for i, ws in enumerate(self.robot.workspaces[1:]):
                    float_Nchannel = max(1.0, exp(self.__scale_Nchannel[i] * ccpar[i]))
                    ws.param.float_Nchannel = float_Nchannel
                    ws.param.modelinfo[0].fastmodel.Nchannel = int(round(float_Nchannel))
    def robot_optimize(self):
        try:
            self.robot.working = True
            self.robot.LL = 0.0
            self.__last_fits_iter = -1
            iterations = 0
            pars = grads = []
            Hessian = None
            self.robot.LL, pars, iterations, grads, Hessian = super(RatesFace, self).robot_optimize()
            self.robot.workspaces[0].param.calc_std(Hessian)
            if self.accel and (self.__last_fits_iter < iterations):
                self.robot_queue_fits(iterations, True)
        except:
            self.robot.send_exception()
        gobject.idle_add(self.__onOptDone, self.robot.LL, iterations, grads, Hessian, self.robot.workspaces, self.robot.model_prep, self.data_prep)
        self.robot.working = False
        return self.robot.LL, pars, iterations, grads, Hessian
    def robot_on_iteration(self, pars, grads, iterations, score):
        if self.robot.workspaces[0].param.modelinfo[0].fastmodel.obj[0].stop_flag:
            raise KeyboardInterrupt()
        else:
            self.robot_set_pars(pars)
            self.robot_queue_fits(iterations)
            self.robot.LL = score
            K0s = [array(mi.fastmodel.K0, copy=True) for mi in self.robot.workspaces[0].param.modelinfo]
            K1s = [array(mi.fastmodel.K1, copy=True) for mi in self.robot.workspaces[0].param.modelinfo]
            K2s = [array(mi.fastmodel.K2, copy=True) for mi in self.robot.workspaces[0].param.modelinfo]
            rest = ([array(mi.fastmodel.amp, copy=True) for mi in self.robot.workspaces[0].param.modelinfo],
                    [numpy.sqrt(array(mi.fastmodel.var, copy=True)) for mi in self.robot.workspaces[0].param.modelinfo],
                    [mi.fastmodel.amp0 for mi in self.robot.workspaces[0].param.modelinfo],
                    [sqrt(mi.fastmodel.var0) for mi in self.robot.workspaces[0].param.modelinfo],
                    [mi.fastmodel.Nchannel for mi in self.robot.workspaces[0].param.modelinfo])
            gobject.idle_add(self.__onIteration, iterations, - score, grads, self.robot.optNchannel,
                             self.robot.workspaces, self.robot.workspaces[0].param.modelinfo,
                             K0s, K1s, K2s, *rest)
    def robot_queue_fits(self, iteration=0, force=False):
        if (not force) and self.optimizing and self.accel:
            if (datetime.datetime.now() - self.__last_fits_time).seconds < MIN_SECONDS_BETWEEN_FIT:
                return
        if (not self.recursive_fit) and hasattr(qubx.global_namespace.QubX.Modeling, 'MacRates'):
            MacRates = qubx.global_namespace.QubX.Modeling.MacRates
            if not MacRates.live:
                MacRates.live = True
                MacRates.live = False
            return
        self.__last_fits_time = datetime.datetime.now()
        self.__last_fits_iter = iteration
        self.fitbot.serial += 1 # shared serial so it updates all files or none
        for data in self.robot.data:
            #print 'Ndata:',self.robot.data.Ndata
            chunks = []
            samples = data.get_buffer()
            Ndata = data.Ndata
            start = 0
            for firsts, lasts, count in data.ff_ll_counts:
                means = numpy.array(samples[start:start+count, 0], copy=True)
                stds = numpy.sqrt(array(samples[start:start+count, 1])) # what if some are neg/nan?
                #means = array([sample.fit for sample in samples[start:start+count]], dtype='float32')
                #stds = array([sqrt(sample.fit_var) for sample in samples[start:start+count]], dtype='float32')
                chunks.append((firsts, lasts, means, stds))
                start += count
            self.fitbot.do(self.fitbot_set_fits, data.file, data.iSignal, chunks, self.fitbot.serial)
    def robot_on_status(self, s):
        gobject.idle_add(self.outStatus.write, s+"\n")
    def fitbot_set_fits(self, datafile, iSignal, chunks, serial):
        # separate thread to wait for gui to finish drawing, discard stale iterations' fit curves, apply latest
        if serial == self.fitbot.serial:
            gobject.idle_add(self.do_set_fits, datafile, iSignal, chunks)
        self.fitbot.last_serial = serial
    def do_set_fits(self, datafile, iSignal, chunks):
        fits = datafile.fits[iSignal].idl
        fits.clear()
        for ff, ll, mm, ss in chunks:
            fits.set_fit(ff, ll, mm, ss)
        datafile.OnChangeFits(datafile, iSignal)


def set_modelprep_rates(mp):
    # override of ModelPrep.set_rates2 for multi_model
    def set_rates(*K012s):
        for prep, K012 in itertools.izip(mp.preps, K012s):
            prep.K0, prep.K1, prep.K2 = K012
            K012ToModel(prep.K0, prep.K1, prep.K2, prep.model)
    return set_rates



## [x] start at equilibrium P           Gleak: [_________0.0_]  [x] live  LL: [___-33952.5_] ##
## [x] resample by delta [___1.0_]      Vleak: [_________0.0_]            [[    Get LL    ]] ##
## |==========================================================| Optimize:                    ##
## |MacroR started...                                       | [x] channel count            ## 
## | 1:      -33129                                           | [x] amp                      ##
## | 2:      -32981                                           | [x] var                      ##
## | 3:      -32360                                           | [x] rates (grayed out; tooltip fix via constraints)
## | 4:      -32110                                           |                              ##
## |==========================================================| [ DFP       |v] [ Optimize ] ##

# repair or lose pascal max*ll; harmonize with QuB  # was: fix reduceconstraints in pascal (what else has changed?)
# G/Vleak sensitive only for IeqFv models?

# other outputs (tabular results?)
# averagecount, variance of current
# fast/slow numgrad?
# useVar always?
